#include <stdio.h>
#include <string.h>
#include "str_utils.h"
#include "messages.h"
#include "array.h"
#include "stg_graph.h"
#include "stg_graph_utils.h"
#include "stg.h"

//---------------------------------------------------------------------------
const char* ASTG_KEYWORD_MODEL         = ".model";
const char* ASTG_KEYWORD_INPUTS        = ".inputs";
const char* ASTG_KEYWORD_OUTPUTS       = ".outputs";
const char* ASTG_KEYWORD_INTERNAL      = ".internal";
const char* ASTG_KEYWORD_DUMMY         = ".dummy";
const char* ASTG_KEYWORD_INITIAL_STATE = ".initial state";
const char* ASTG_KEYWORD_GRAPH         = ".graph";
const char* ASTG_KEYWORD_MARKING       = ".marking";
const char* ASTG_KEYWORD_CAPACITY      = ".capacity";
const char* ASTG_KEYWORD_END           = ".end";

const char* ASTG_PATTERN_HIGH          = "%s";
const char* ASTG_PATTERN_LOW           = "!%s";
const char* ASTG_SYMBOLS_POLARITY      = "+-";

const char* PATTERN_PLACE              = "%s%d";
const char* NAME_PLACE                 = "p";

//---------------------------------------------------------------------------
Graph::Graph(const char* _name, Tag _tag, void* _data) 
: Object(_name, _tag, _data), model(NULL)
{
  signal_hashset = new SignalHashSet();
  signals = new SignalArray();
  
  transition_hashset = new TransitionHashSet();
  transitions = new TransitionArray();

  place_hashset = new PlaceHashSet();
  places = new PlaceArray();
}

//---------------------------------------------------------------------------
Graph::Graph(const Graph& graph) 
: Object(graph), model(NULL)
{
  setModel(graph.model);
  
  signal_hashset = new SignalHashSet();
  signals = new SignalArray(graph.getSignalCount());
  for (int i = 0; i < graph.getSignalCount(); i++)
    addSignal( new Signal(*graph.getSignal(i)) );
  
  transition_hashset = new TransitionHashSet();
  transitions = new TransitionArray(graph.getTransitionCount());
  for (int i = 0; i < graph.getTransitionCount(); i++)
    addTransition( new Transition(*graph.getTransition(i)) );
  
  place_hashset = new PlaceHashSet();
  places = new PlaceArray(graph.getPlaceCount());
  for (int i = 0; i < graph.getPlaceCount(); i++)
    addPlace( new Place(*graph.getPlace(i)) );
  
  calcIncidence();
}

//---------------------------------------------------------------------------
void Graph::calcIncidence()
{
  for (int i = 0; i < getTransitionCount(); i++)
    calcTransitionIncidence(getTransition(i));
  for (int i = 0; i < getPlaceCount(); i++)
    calcPlaceIncidence(getPlace(i));
}

//---------------------------------------------------------------------------
void Graph::calcTransitionIncidence(Transition* transition)
{
  for (int i = 0; i < transition->getPresetCount(); i++)
  {
    Place* place_cur = findPlace( transition->getPreset(i)->getName() );
    transition->replacePreset(place_cur, i);
  }
  for (int i = 0; i < transition->getPostsetCount(); i++)
  {
    Place* place_cur = findPlace( transition->getPostset(i)->getName() );
    transition->replacePostset(place_cur, i);
  }
  for (int i = 0; i < transition->getReadsetCount(); i++)
  {
    Place* place_cur = findPlace( transition->getReadset(i)->getName() );
    transition->replaceReadset(place_cur, i);
  }
}
 
//---------------------------------------------------------------------------
void Graph::calcPlaceIncidence(Place* place)
{
  for (int i = 0; i < place->getPresetCount(); i++)
  {
    Transition* transition_cur = findTransition( place->getPreset(i)->getName() );
    place->replacePreset(transition_cur, i);
  }
  for (int i = 0; i < place->getPostsetCount(); i++)
  {
    Transition* transition_cur = findTransition( place->getPostset(i)->getName() );
    place->replacePostset(transition_cur, i);
  }
  for (int i = 0; i < place->getReadsetCount(); i++)
  {
    Transition* transition_cur = findTransition( place->getReadset(i)->getName() );
    place->replaceReadset(transition_cur, i);
  }
}

//---------------------------------------------------------------------------
Graph::~Graph()
{
  clean();
  
  delete [] model;
  
  delete signal_hashset; signal_hashset = NULL;
  delete signals; signals = NULL;
  
  delete transition_hashset; transition_hashset = NULL;
  delete transitions; transitions = NULL;

  delete place_hashset; place_hashset = NULL;
  delete places; places = NULL;
}

//---------------------------------------------------------------------------
void Graph::clean()
{
  if (signals) 
    signals->clean(true);
  if (transitions) 
    transitions->clean(true);
  if (places) 
    places->clean(true);
}

//---------------------------------------------------------------------------
Signal* Graph::addSignal(const char* signal_name, Signal::Type signal_type)
{
  Signal* signal_new = new Signal(signal_name, signal_type);
  Signal* signal = signal_hashset->find(signal_new);
  if (signal)
  	delete signal_new;
  else
  {
  	signal = signal_new;
   	signals->add(signal);
   	signal_hashset->add(signal);
  }
  return signal;
}

//---------------------------------------------------------------------------
bool Graph::addSignal(Signal* signal, int position)
{
  bool result = false;
  if ( !signal_hashset->find(signal) )
  {
    if (position < 0)
     	signals->add(signal);
    else
      signals->insert(signal, position);
   	signal_hashset->add(signal);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void Graph::addSignals(SignalArray* _signals)
{
  if (_signals)
  for (int i = 0; i < _signals->count(); i++)
  {
  	Signal* signal_cur = (*_signals)[i];
  	if ( !addSignal(signal_cur) )
      delete signal_cur;
  }
}

//---------------------------------------------------------------------------
void Graph::removeSignal(Signal* signal)
{
  signal_hashset->remove(signal);
  signals->remove(signal, true);
}

//---------------------------------------------------------------------------
void Graph::removeSignals(SignalArray* _signals)
{
  if (_signals)
  {
    SignalArray* signals_tmp = new SignalArray(*_signals);
    for (int i = 0; i < signals_tmp->count(); i++)
    {
      Signal* signal_cur = (*signals_tmp)[i];
      removeSignal(signal_cur);
    }
    delete signals_tmp;
  }
}

//---------------------------------------------------------------------------
Signal* Graph::findSignal(const char* signal_name) const
{
  Signal* signal_mask = new Signal(signal_name);
  Signal* signal_cur = signal_hashset->find(signal_mask);
  delete signal_mask;
  return signal_cur;
}

//---------------------------------------------------------------------------
Signal* Graph::findSignalSmart(const char* signal_name_ext) const
{
  char* signal_name = namecopy(signal_name_ext);
  char* suffix_str = strpbrk(signal_name, "+-/");
  if (suffix_str)
  {
    suffix_str[0] = '\0';
  }
  Signal* signal_mask = new Signal(signal_name);
  delete [] signal_name;
  Signal* signal_cur = signal_hashset->find(signal_mask);
  delete signal_mask;
  return signal_cur;
}

//---------------------------------------------------------------------------
void Graph::renameSignal(Signal* signal, const char* signal_name)
{
  signal_hashset->remove(signal);
  signal->setName(signal_name);
  signal_hashset->add(signal);
}

//---------------------------------------------------------------------------
Transition* Graph::addTransition(const char* transition_name)
{
  Transition* transition_new = new Transition(transition_name);
  Transition* transition = transition_hashset->find(transition_new);
  if (transition)
  	delete transition_new;
  else
  {
  	transition = transition_new;
   	transitions->add(transition);
   	transition_hashset->add(transition);
  }
  return transition;
}

//---------------------------------------------------------------------------
bool Graph::addTransition(Transition* transition, int position)
{
  bool result = false;
  if ( !transition_hashset->find(transition) )
  {
    if (position < 0)
     	transitions->add(transition);
    else
      transitions->insert(transition, position);
   	transition_hashset->add(transition);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void Graph::addTransitions(TransitionArray* _transitions)
{
  if (_transitions)
  for (int i = 0; i < _transitions->count(); i++)
  {
  	Transition* transition_cur = (*_transitions)[i];
  	if ( !addTransition(transition_cur) )
      delete transition_cur;
  }
}

//---------------------------------------------------------------------------
void Graph::removeTransition(Transition* transition)
{
  removeConsumingArcs(transition->getPreset(), transition);
  removeProducingArcs(transition, transition->getPostset());
  removeReadArcs(transition->getReadset(), transition);
  transition_hashset->remove(transition);
  transitions->remove(transition, true);
}

//---------------------------------------------------------------------------
void Graph::removeTransitions(TransitionArray* _transitions)
{
  if (_transitions)
  {
    TransitionArray* transitions_tmp = new TransitionArray(*_transitions);
    for (int i = 0; i < transitions_tmp->count(); i++)
    {
      Transition* transition_cur = (*transitions_tmp)[i];
      removeTransition(transition_cur);
    }
    delete transitions_tmp;
  }
}

//---------------------------------------------------------------------------
Transition* Graph::findTransition(const char* transition_name) const
{
  Transition* transition_mask = new Transition(transition_name);
  Transition* transition_cur = transition_hashset->find(transition_mask);
  delete transition_mask;
  return transition_cur;
}

//---------------------------------------------------------------------------
void Graph::renameTransition(Transition* transition, const char* transition_name)
{
  transition_hashset->remove(transition);
  transition->setName(transition_name);
  transition_hashset->add(transition);
}

//---------------------------------------------------------------------------
Place* Graph::addPlace(const char* place_name)
{
  Place* place_new = new Place(place_name);
  Place* place = place_hashset->find(place_new);
  if (place)
  {
    if (place->testTag(TAG_GENERATED))
    {
      char* place_name_new = genPlaceName();
      place_hashset->remove(place);
      place->setName(place_name_new);
      place_hashset->add(place);
      place = NULL;
    }
  }
  if (place)
    delete place_new;
  else
  {
  	place = place_new;
   	places->add(place);
   	place_hashset->add(place);
  }
  return place;
}

//---------------------------------------------------------------------------
bool Graph::addPlace(Place* place, int position)
{
  bool result = false;
  if (!place_hashset->find(place))
  {
    if (position < 0)
      places->add(place);
    else
      places->insert(place, position);
   	place_hashset->add(place);
   	result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void Graph::addPlaces(PlaceArray* _places)
{
  if (_places)
  for (int i = 0; i < _places->count(); i++)
  {
  	Place* place_cur = (*_places)[i];
  	if (!addPlace(place_cur))
      delete place_cur;
  }
}

//---------------------------------------------------------------------------
void Graph::removePlace(Place* place)
{
  removeConsumingArcs(place, place->getPostset());
  removeProducingArcs(place->getPreset(), place);
  removeReadArcs(place, place->getReadset());
  place_hashset->remove(place);
  places->remove(place, true);
}

//---------------------------------------------------------------------------
void Graph::removePlaces(PlaceArray* _places)
{
  if (places)
  {
    PlaceArray* places_tmp = new PlaceArray(*_places);
    for (int i = 0; i < places_tmp->count(); i++)
    {
      Place* place_cur = (*places_tmp)[i];
      removePlace(place_cur);
    }
    delete places_tmp;
  }
}

//---------------------------------------------------------------------------
Place* Graph::findPlace(const char* place_name) const
{
  Place* place_mask = new Place(place_name);
  Place* place_cur = place_hashset->find(place_mask);
  delete place_mask;
  return place_cur;
}


//---------------------------------------------------------------------------
Place* Graph::findPlace(const char* transition_name_preset, const char* transition_name_postset) const
{
  Place* place = NULL;
  Transition* transition_preset = findTransition(transition_name_preset);
  Transition* transition_postset = findTransition(transition_name_postset);
  for (int i = 0; i < getPlaceCount(); i++)
  {
    Place* place_cur = getPlace(i);
    if ( isProducingArc(transition_preset, place_cur) 
      && isConsumingArc(place_cur, transition_postset) )
      place = place_cur;
  }
  return place;
}

//---------------------------------------------------------------------------
char* Graph::genPlaceName()
{
  int i = 0;
  char* name_tmp = name2pattern(PATTERN_PLACE, NAME_PLACE, i);
  while ( findPlace(name_tmp) )
  {
    delete [] name_tmp;
    i++;
    name_tmp = name2pattern(PATTERN_PLACE, NAME_PLACE, i);
  }
  return name_tmp;
}

//---------------------------------------------------------------------------
void Graph::renamePlace(Place* place, const char* place_name)
{
  place_hashset->remove(place);
  place->setName(place_name);
  place_hashset->add(place);
}

//---------------------------------------------------------------------------
void Graph::addArc(Object* node_from, Object* node_to)
{
  Signal* signal_from = findSignalSmart(node_from->getName());
  Signal* signal_to = findSignalSmart(node_to->getName());
  if (signal_from && signal_to)
  {
    Place* place_middle = addPlace( genPlaceName() );
    place_middle->addTag(TAG_GENERATED);
    Transition* transition_from = addTransition(node_from->getName());
    Transition* transition_to = addTransition(node_to->getName());
    addProducingArc(transition_from, place_middle);
    addConsumingArc(place_middle, transition_to);
  }
  else
  if (signal_from)
  {
  	// transition -> place
    Transition* transition = addTransition(node_from->getName());
    Place* place = addPlace(node_to->getName());
    if ( isConsumingArc(place, transition) )
    {
      removeConsumingArc(place, transition);
      if ( ! isReadArc(place, transition) )
        addReadArc(place, transition);
    }
    else
      addProducingArc(transition, place);
  }
  else
  if (signal_to)
  {
  	// place -> transition
    Place* place = addPlace(node_from->getName());
    Transition* transition = addTransition(node_to->getName());
    if ( isProducingArc(transition, place) )
    {
      removeProducingArc(transition, place);
      if ( ! isReadArc(place, transition) )
        addReadArc(place, transition);
    }
    else
      addConsumingArc(place, transition);
  }
}

//---------------------------------------------------------------------------
void Graph::readAstg(const char* file_names)
{
  if (file_names)
  {
    int name_index = 0;
    while (char* file_name = names2name(file_names, name_index++))
    {
      if (strcmp(file_name, NAME_STDIO) == 0)
        stg_in = stdin;
      else
        if ((stg_in = fopen(file_name, "r")) == NULL)
          Error("opening input file <%s>", file_name);
      stg_parse(this, file_name);
      if (stg_in != stdin)
        fclose(stg_in);
      delete [] file_name;
    }
  }
}

//---------------------------------------------------------------------------
void Graph::writeAstgSignals(FILE* file, Signal::Type type, const char* label)
{
  bool label_printed = false;
  for (int i = 0; i < getSignalCount(); i++)
  {
    Signal* signal_cur = getSignal(i);
    if (signal_cur->getType() == type)
    {
      if (!label_printed)
      {
        fprintf(file, label);
        label_printed = true;
      }
      fprintf(file, " %s", signal_cur->getName());
    }
  }
  if (label_printed)
    fprintf(file, "\n");
}

//---------------------------------------------------------------------------
void Graph::writeAstgInitialStates(FILE* file)
{
  bool label_printed = false;
  for (int i = 0; i < getSignalCount(); i++)
  {
  	Signal* signal_cur = getSignal(i);
    if ( (signal_cur->getState() == Signal::STATE_LOW)
      || (signal_cur->getState() == Signal::STATE_HIGH) )
    {
      if (!label_printed)
      {
        fprintf(file, ASTG_KEYWORD_INITIAL_STATE);
        label_printed = true;
      }
      fprintf(file, " ");
      if (signal_cur->getState() == Signal::STATE_LOW)
        fprintf( file, ASTG_PATTERN_LOW, signal_cur->getName() );
      if (signal_cur->getState() == Signal::STATE_HIGH)
        fprintf( file, ASTG_PATTERN_HIGH, signal_cur->getName() );
    }
  }
  if (label_printed)
    fprintf(file, "\n");
}

//---------------------------------------------------------------------------
void Graph::writeAstgProducingArcs(FILE* file, bool show_implicit)
{
  for (int i = 0; i < getTransitionCount(); i++)
  {
    Transition* transition_cur = getTransition(i);
    if (transition_cur->getPostsetCount() > 0)
    {
      fprintf(file, "%s", transition_cur->getName());
      for (int j = 0; j < transition_cur->getPostsetCount(); j++)
      {
        Place* place_cur = transition_cur->getPostset(j);
        if ( show_implicit || !place_cur->isImplicit() )
          fprintf(file, " %s", place_cur->getName() );
        else  
          fprintf(file, " %s", place_cur->getPostset(0)->getName());
      }
      fprintf(file, "\n");
    }
  }
}

//---------------------------------------------------------------------------
void Graph::writeAstgConsumingArcs(FILE* file, bool show_implicit)
{
  for (int i = 0; i < getPlaceCount(); i++)
  {
    Place* place_cur = getPlace(i);
    if (place_cur->getPostsetCount() > 0)
    {
      if ( show_implicit || !place_cur->isImplicit() )
      {
        fprintf(file, "%s", place_cur->getName());
        for (int j = 0; j < place_cur->getPostsetCount(); j++)
        {
          Transition* transition_cur = place_cur->getPostset(j);
          fprintf(file, " %s", transition_cur->getName());
        }
        fprintf(file, "\n");
      }
    }
  }
}

//---------------------------------------------------------------------------
void Graph::writeAstgReadArcs(FILE* file)
{
  for (int i = 0; i < getPlaceCount(); i++)
  {
    Place* place_cur = getPlace(i);
    if (place_cur->getReadsetCount() > 0)
    {
      for (int j = 0; j < place_cur->getReadsetCount(); j++)
      {
        Transition* transition_cur = place_cur->getReadset(j);
        fprintf(file, "%s %s\n", place_cur->getName(), transition_cur->getName());
        fprintf(file, "%s %s\n", transition_cur->getName(), place_cur->getName());
      }
    }
  }
}

//---------------------------------------------------------------------------
void Graph::writeAstgArcs(FILE* file, bool show_implicit)
{
  fprintf(file, "%s\n", ASTG_KEYWORD_GRAPH);
  writeAstgProducingArcs(file, show_implicit);
  writeAstgConsumingArcs(file, show_implicit);
  writeAstgReadArcs(file);
}

//---------------------------------------------------------------------------
void Graph::writeAstgMarking(FILE* file, bool show_implicit)
{
  fprintf(file, "%s {", ASTG_KEYWORD_MARKING);
  bool label_printed = false;
  for (int i = 0; i < getPlaceCount(); i++)
  {
    Place* place_cur = getPlace(i);
    if (place_cur->getMarking() > 0)
    {
      if (label_printed)
        fprintf(file, " ");
      if ( show_implicit || !place_cur->isImplicit() )
        fprintf(file, "%s", place_cur->getName());
      else
        fprintf(file, "<%s, %s>", place_cur->getPreset(0)->getName(), place_cur->getPostset(0)->getName());
      label_printed = true;
    }
  }
  fprintf(file, "}\n");
}

//---------------------------------------------------------------------------
void Graph::writeAstg(const char* file_name, bool show_implicit)
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    // model
    if (model)
      fprintf(file, "%s %s\n", ASTG_KEYWORD_MODEL, model);
    // signals
    writeAstgSignals(file, Signal::TYPE_INPUT, ASTG_KEYWORD_INPUTS);
    writeAstgSignals(file, Signal::TYPE_OUTPUT, ASTG_KEYWORD_OUTPUTS);
    writeAstgSignals(file, Signal::TYPE_INTERNAL, ASTG_KEYWORD_INTERNAL);
    writeAstgSignals(file, Signal::TYPE_DUMMY, ASTG_KEYWORD_DUMMY);
    // initial states
    writeAstgInitialStates(file);
    // graph
    writeAstgArcs(file, show_implicit);
    // marking
    writeAstgMarking(file, show_implicit);
    // end
    fprintf(file, "%s\n", ASTG_KEYWORD_END);
    if (file != stdout)
      fclose(file);
  }
}
