#ifndef _STG_GRAPH_H_
#define _STG_GRAPH_H_

#include <stdio.h>
#include "str_utils.h"
#include "array.h"
#include "hashset.h"
#include "stg_object.h"
#include "stg_signal.h"
#include "stg_place.h"
#include "stg_transition.h"

//---------------------------------------------------------------------------
extern const char* ASTG_KEYWORD_MODEL;
extern const char* ASTG_KEYWORD_INPUTS;
extern const char* ASTG_KEYWORD_OUTPUTS;
extern const char* ASTG_KEYWORD_INTERNAL;
extern const char* ASTG_KEYWORD_DUMMY;
extern const char* ASTG_KEYWORD_INITIAL_STATE;
extern const char* ASTG_KEYWORD_GRAPH;
extern const char* ASTG_KEYWORD_MARKING;
extern const char* ASTG_KEYWORD_CAPACITY;
extern const char* ASTG_KEYWORD_END;
extern const char* ASTG_PATTERN_HIGH;
extern const char* ASTG_PATTERN_LOW;
extern const char* ASTG_SYMBOLS_POLARITY;

//---------------------------------------------------------------------------
class Graph;
typedef Array<Graph, ObjectCompareName> GraphArray;
typedef HashSet<Graph, ObjectHashName, ObjectEqualName> GraphHashSet;

//---------------------------------------------------------------------------
class Graph : public Object
{
  public:
    Graph(const char* _name = NULL, Tag _tag = TAG_UNDEFINED, void* _data = NULL);
    Graph(const Graph& graph);
    ~Graph();
	
    void setModel(const char* _model);
    const char* getModel() const;

    int getSignalCount() const;
    Signal* getSignal(const int signal_index) const;
    SignalArray* getSignals() const;
    Signal* addSignal(const char* signal_name, Signal::Type signal_type = Signal::TYPE_UNDEFINED);
    bool addSignal(Signal* signal, int position = -1);
    void addSignals(SignalArray* _signals);
    void removeSignal(Signal* signal);
    void removeSignals(SignalArray* _signals);
    Signal* findSignal(const char* signal_name) const;
    Signal* findSignalSmart(const char* signal_name_ext) const;
    void renameSignal(Signal* signal, const char* signal_name);
    
    int getTransitionCount() const;
    Transition* getTransition(const int transition_index) const;
    TransitionArray* getTransitions() const;
    Transition* addTransition(const char* transition_name);
    bool addTransition(Transition* transition, int position = -1);
    void addTransitions(TransitionArray* _transitions);
    void removeTransition(Transition* transition);
    void removeTransitions(TransitionArray* _transitions);
    Transition* findTransition(const char* transition_name) const;
    void renameTransition(Transition* transition, const char* transition_name);

    int getPlaceCount() const;
    Place* getPlace(const int place_index) const;
    PlaceArray* getPlaces() const;
    Place* addPlace(const char* place_name);
    bool addPlace(Place* place, int position = -1);
    void addPlaces(PlaceArray* _places);
    void removePlace(Place* place);
    void removePlaces(PlaceArray* _places);
    Place* findPlace(const char* place_name) const;
    Place* findPlace(const char* transition_name_preset, const char* transition_name_postset) const;
    void renamePlace(Place* place, const char* place_name);

    void addArc(Object* node_from, Object* node_to);

    void readAstg(const char* file_names);
    void writeAstg(const char* file_name, bool show_implicit = false);

  protected:
    char* model;
    SignalArray* signals;
    SignalHashSet* signal_hashset;
    TransitionArray* transitions;
    TransitionHashSet* transition_hashset;
    PlaceArray* places;
    PlaceHashSet* place_hashset;

    void clean();
    char* genPlaceName();
    void calcIncidence();
    void calcTransitionIncidence(Transition* transition);
    void calcPlaceIncidence(Place* place);

    void writeAstgSignals(FILE* file, Signal::Type type, const char* label);
    void writeAstgInitialStates(FILE* file);
    void writeAstgProducingArcs(FILE* file, bool show_implicit);
    void writeAstgConsumingArcs(FILE* file, bool show_implicit);
    void writeAstgReadArcs(FILE* file);
    void writeAstgArcs(FILE* file, bool show_implicit);
    void writeAstgMarking(FILE* file, bool show_implicit);
};

//---------------------------------------------------------------------------
inline void Graph::setModel(const char* _model)
{
  delete [] model;
  model = namecopy(_model);
}

//---------------------------------------------------------------------------
inline const char* Graph::getModel() const
{
  return model;
}

//---------------------------------------------------------------------------
inline int Graph::getSignalCount() const
{
  return signals->count();
}

//---------------------------------------------------------------------------
inline Signal* Graph::getSignal(int signal_index) const
{
  return (*signals)[signal_index];
}

//---------------------------------------------------------------------------
inline SignalArray* Graph::getSignals() const
{
  return signals;
}

//---------------------------------------------------------------------------
inline int Graph::getTransitionCount() const
{
  return transitions->count();
}

//---------------------------------------------------------------------------
inline Transition* Graph::getTransition(int transition_index) const
{
  return (*transitions)[transition_index];
}

//---------------------------------------------------------------------------
inline TransitionArray* Graph::getTransitions() const
{
  return transitions;
}

//---------------------------------------------------------------------------
inline int Graph::getPlaceCount() const
{
  return places->count();
}

//---------------------------------------------------------------------------
inline Place* Graph::getPlace(int place_index) const
{
  return (*places)[place_index];
}

//---------------------------------------------------------------------------
inline PlaceArray* Graph::getPlaces() const
{
  return places;
}

#endif
