#include <stdio.h>
#include "messages.h"
#include "stg_object.h"
#include "stg_transition.h"
#include "stg_place.h"
#include "stg_cluster.h"
#include "stg_graph.h"
#include "stg_graph_cluster.h"
#include "stg_notation.h"

//---------------------------------------------------------------------------
const char* PATTERN_CLUSTER            = "%s%d";
const char* NAME_CLUSTER_TRACKER       = "tracker_";
const char* NAME_CLUSTER_UNDEFINED     = "cluster_";

//---------------------------------------------------------------------------
GraphCluster::GraphCluster(const char* _name, Tag _tag, void* _data)
: Graph(_name, _tag, _data)
{
  cluster_hashset = new ClusterHashSet();
  clusters = new ClusterArray(2);
}

//---------------------------------------------------------------------------
GraphCluster::GraphCluster(const GraphCluster& graph_cluster)
: Graph(graph_cluster)
{
  clusters = new ClusterArray(graph_cluster.clusters->count());
  for (int i = 0; i < graph_cluster.clusters->count(); i++)
  {
    Cluster* cluster_cur = new Cluster(*(*graph_cluster.clusters)[i]);
    calcClusterIncidence(cluster_cur);
    addCluster(cluster_cur);
  }
}

//---------------------------------------------------------------------------
GraphCluster::~GraphCluster()
{
  clusters->clean(true);
  delete clusters; clusters = NULL;
}

//---------------------------------------------------------------------------
char* GraphCluster::guessClusterName(Cluster* cluster)
{
  char* cluster_name = NULL;
  if (cluster->getTransitionCount() > 0)
  {
    const char* transition_name = cluster->getTransition(0)->getName();
    int len_protocol = 0;
    int len_prefix = strlen(NOTATION_PROTOCOL_DUMMY_PREFIX);
    int len_suffix = strlen(NOTATION_PROTOCOL_DUMMY_SUFFIX);
    if ( strncmp(transition_name, NOTATION_PROTOCOL_DUMMY_PREFIX, len_prefix) == 0 )
    {
      char* str_suffix = strstr(&transition_name[len_prefix], NOTATION_PROTOCOL_DUMMY_SUFFIX);
      if (str_suffix)
      {
        len_protocol = str_suffix - transition_name + len_suffix;
        for (int transition_index = 1; transition_index < cluster->getTransitionCount(); transition_index++)
        {
          Transition* transition_cur = cluster->getTransition(transition_index);
          if ( strncmp(transition_cur->getName(), transition_name, len_protocol) != 0 )
          {
            len_protocol = 0;
            break;
          }
        }
        if (len_protocol > 0)
        {
           int len_cluster = len_protocol - len_prefix - len_suffix;
           cluster_name = new char[len_cluster + 1];
           strncpy(cluster_name, &transition_name[len_prefix], len_cluster);
           cluster_name[len_cluster] = '\0';
        }
      }
    }
  }
  return cluster_name;
}

//---------------------------------------------------------------------------
char* GraphCluster::genClusterName(const char* cluster_name)
{
  int i = 0;
  char* name_tmp = namecopy(cluster_name);
  while ( findCluster(name_tmp) )
  {
    delete [] name_tmp;
    name_tmp = name2pattern(PATTERN_CLUSTER, cluster_name, i);
    i++;
  }
  return name_tmp;
}

//---------------------------------------------------------------------------
void GraphCluster::calcClusterIncidence(Cluster* cluster)
{
  for (int i = 0; i< cluster->getTransitionCount(); i++)
  {
    Transition* transition_cur = findTransition( cluster->getTransition(i)->getName() );
    cluster->replaceTransition(transition_cur, i);
  }
  for (int i = 0; i< cluster->getPlaceCount(); i++)
  {
    Place* place_cur = findPlace( cluster->getPlace(i)->getName() );
    cluster->replacePlace(place_cur, i);
  }
}

//---------------------------------------------------------------------------
bool GraphCluster::addCluster(Cluster* cluster)
{
  bool result = false;
  if ( !cluster_hashset->find(cluster) )
  {
    clusters->add(cluster);
    cluster_hashset->add(cluster);
    result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void GraphCluster::removeCluster(Cluster* cluster)
{
  cluster_hashset->remove(cluster);
  clusters->remove(cluster, true);
}

//---------------------------------------------------------------------------
Cluster* GraphCluster::findCluster(const char* cluster_name) const
{
  Cluster* cluster_mask = new Cluster(cluster_name);
  Cluster* cluster_cur = cluster_hashset->find(cluster_mask);
  delete cluster_mask;
  return cluster_cur;
}

//---------------------------------------------------------------------------
Cluster* GraphCluster::findCluster(Object* object) const
{
  for (int cluster_index = 0; cluster_index < clusters->count(); cluster_index++)
  {
    Cluster* cluster_cur = (*clusters)[cluster_index];
    if ( cluster_cur->locatePlace((Place*)object) >= 0 )
      return cluster_cur;
    if ( cluster_cur->locateTransition((Transition*)object) >= 0 )
      return cluster_cur;
  }
  return NULL;
}

//---------------------------------------------------------------------------
void GraphCluster::findTransitionCluster(Cluster* cluster, Transition* transition)
{
  if ( !transition->testTag(TAG_PROCESSED) )
  {
    transition->addTag(TAG_PROCESSED);
    cluster->addTransition(transition);
    for (int preset_index = 0; preset_index < transition->getPresetCount(); preset_index++)
    {
      Place* place_cur = transition->getPreset(preset_index);
      Text(4, "adding place <%s> (preset of transition <%s>)\n", place_cur->getName(), transition->getName());
      findPlaceCluster(cluster, place_cur);
    }
    for (int postset_index = 0; postset_index < transition->getPostsetCount(); postset_index++)
    {
      Place* place_cur = transition->getPostset(postset_index);
      Text(4, "adding place <%s> (postset of transition <%s>)\n", place_cur->getName(), transition->getName());
      findPlaceCluster(cluster, place_cur);
    }
  }
}

//---------------------------------------------------------------------------
void GraphCluster::findPlaceCluster(Cluster* cluster, Place* place)
{
  if (!place->testTag(TAG_PROCESSED))
  {
    place->addTag(TAG_PROCESSED);
    cluster->addPlace(place);
    for (int preset_index = 0; preset_index < place->getPresetCount(); preset_index++)
    {
      Transition* transition_cur = place->getPreset(preset_index);
      Text(4, "adding transistion <%s> (preset of place <%s>)\n", transition_cur->getName(), place->getName());
      findTransitionCluster(cluster, transition_cur);
    }
    for (int postset_index = 0; postset_index < place->getPostsetCount(); postset_index++)
    {
      Transition* transition_cur = place->getPostset(postset_index);
      Text(4, "adding transistion <%s> (postset of place <%s>)\n", transition_cur->getName(), place->getName());
      findTransitionCluster(cluster, transition_cur);
    }
  }
}

//---------------------------------------------------------------------------
void GraphCluster::calcClusterType(Cluster* cluster)
{
  // cheack a single-signal cluster
  Signal* signal_single = NULL;
  if ( (cluster->getPlaceCount() == 2) && (cluster->getTransitionCount() > 0) )
  {
    signal_single = findSignal( cluster->getTransition(0)->getSignalName() );
    for (int i = 1; signal_single && ( i < cluster->getTransitionCount() ); i++)
    {
      if ( signal_single != findSignal( cluster->getTransition(i)->getSignalName() ) )
        signal_single = NULL;
    }
  }
  // count dummies
  int dummy_count = 0;
  for (int transition_index = 0; transition_index < cluster->getTransitionCount(); transition_index++)
  {
    Transition* transition_cur = cluster->getTransition(transition_index);
    Signal* signal_cur = findSignal(transition_cur->getSignalName());
    if (signal_cur->getType() == Signal::TYPE_DUMMY)
      dummy_count++;
  }
  // tracker
  if (dummy_count == cluster->getTransitionCount())
  {
    cluster->setType(Cluster::TYPE_TRACKER);
    char* cluster_name_guess = guessClusterName(cluster);
    if (cluster_name_guess == NULL)
      cluster_name_guess = namecopy(NAME_CLUSTER_TRACKER);
    char* cluster_name = genClusterName(cluster_name_guess);
    cluster->setName(cluster_name);
    delete [] cluster_name;
    delete [] cluster_name_guess;
  }  
  else
  // undefined
  if (signal_single == NULL)
  {
    cluster->setType(Cluster::TYPE_UNDEFINED);
    char* cluster_name = genClusterName(NAME_CLUSTER_UNDEFINED);
    cluster->setName(cluster_name);
    delete [] cluster_name;
  }
  else
  // single signal
  {
    cluster->setName(signal_single->getName());
    switch (signal_single->getType())
    {
      case Signal::TYPE_INPUT: 
        cluster->setType(Cluster::TYPE_INPUT);
        break;
      case Signal::TYPE_OUTPUT: 
        cluster->setType(Cluster::TYPE_OUTPUT);
        break;
      case Signal::TYPE_INTERNAL: 
        cluster->setType(Cluster::TYPE_INTERNAL);
        break;
      default:
        cluster->setType(Cluster::TYPE_UNDEFINED);
    }
  }
}

//---------------------------------------------------------------------------
void GraphCluster::calcClusters()
{
  clusters->clean(true);

  for (int place_index = 0; place_index < places->count(); place_index++)
    (*places)[place_index]->removeTag(TAG_PROCESSED);
  for (int transition_index = 0; transition_index < transitions->count(); transition_index++)
    (*transitions)[transition_index]->removeTag(TAG_PROCESSED);
  
  for (int place_index = 0; place_index < places->count(); place_index++)
  {
    Place* place_cur = (*places)[place_index];
    if ( !place_cur->testTag(TAG_PROCESSED) )
    {
      Cluster* cluster_new = new Cluster();
      Text(2, "building cluster for place <%s>\n", place_cur->getName());
      findPlaceCluster(cluster_new, place_cur);
      calcClusterType(cluster_new);
      addCluster(cluster_new);
    }
  }
  for (int transition_index = 0; transition_index < transitions->count(); transition_index++)
  {
    Transition* transition_cur = (*transitions)[transition_index];
    if ( !transition_cur->testTag(TAG_PROCESSED) )
    {
      Cluster* cluster_new = new Cluster();
      Text(2, "building cluster for transition <%s>", transition_cur->getName());
      findTransitionCluster(cluster_new, transition_cur);
      calcClusterType(cluster_new);
      addCluster(cluster_new);
    }
  }
}  
