#include <stdio.h>
#include "messages.h"
#include "str_utils.h"
#include "stg_object.h"
#include "stg_object_utils.h"
#include "stg_transition.h"
#include "stg_place.h"
#include "stg_notation.h"
#include "stg_protocol.h"
#include "stg_graph.h"
#include "stg_graph_utils.h"
#include "stg_graph_protode.h"
#include "protocol.h"


//---------------------------------------------------------------------------
const char* PROTOCOL_DEV_NAME = "dev";
const char* PROTOCOL_ENV_NAME = "env";
const char* PROTOCOL_MISC_NAME = "misc";
const char* INTERFACE_NAME = "interface";

const char* ASTG_PROTOCOL_PATTERN_HIGH          = "%s%s";
const char* ASTG_PROTOCOL_PATTERN_LOW           = "!%s%s";
const char* ASTG_PROTOCOL_PATTERN_PLACE         = "%s_%s";

const Tag TAG_PLACE_REDUNDANT            = TAG_BIT_USER_1;
const Tag TAG_PLACE_MANDATORY            = TAG_BIT_USER_2;
const Tag TAG_TRANSITION_PROTOCOL_INPUT  = TAG_BIT_USER_1;
const Tag TAG_TRANSITION_PROTOCOL_OUTPUT = TAG_BIT_USER_2;

//---------------------------------------------------------------------------
GraphProtode::GraphProtode(const char* _name, Tag _tag, void* _data)
: Graph(_name, _tag, _data)
{
  interface = new Graph(INTERFACE_NAME, TAG_UNDEFINED, NULL);
  tracker_hashset = new GraphHashSet();
  trackers = new GraphArray(2);
  protocol_hashset = new ProtocolHashSet();
  protocols = new ProtocolArray(2);
}

//---------------------------------------------------------------------------
GraphProtode::GraphProtode(const GraphProtode& graph_protode)
: Graph(graph_protode)
{
  interface = new Graph(*graph_protode.interface);

  tracker_hashset = new GraphHashSet();
  trackers = new GraphArray(graph_protode.getTrackerCount());
  for (int i = 0; i < graph_protode.getTrackerCount(); i++)
    addTracker( new Graph(*graph_protode.getTracker(i)) );

  protocol_hashset = new ProtocolHashSet();
  protocols = new ProtocolArray(graph_protode.getProtocolCount());
  for (int i = 0; i < graph_protode.getProtocolCount(); i++)
    addProtocol( new Protocol(*graph_protode.getProtocol(i)) );
}

//---------------------------------------------------------------------------
GraphProtode::~GraphProtode()
{
  clean();
  delete interface; interface = NULL;
  delete tracker_hashset; tracker_hashset = NULL;
  delete trackers; trackers = NULL;
  delete protocol_hashset; protocol_hashset = NULL;
  delete protocols; protocols = NULL;
}

//---------------------------------------------------------------------------
void GraphProtode::clean()
{
  protocols->clean(true);
  trackers->clean(true);
}

//---------------------------------------------------------------------------
char* GraphProtode::genSignalDummyName(Transition* transition)
{
  char* signal_dummy_name = NULL;
  if (transition->getPolarity() == Transition::POLARITY_RISING)
    signal_dummy_name = name2pattern("%s%s", transition->getSignalName(), NOTATION_EXPOSED_DUMMY_SUFFIX_RISING);
  else
  if (transition->getPolarity() == Transition::POLARITY_FALLING)
    signal_dummy_name = name2pattern("%s%s", transition->getSignalName(), NOTATION_EXPOSED_DUMMY_SUFFIX_FALLING);
  return signal_dummy_name;
}

//---------------------------------------------------------------------------
char* GraphProtode::genPlaceDupName(Place* place, int dup)
{
  char* place_dup_name = NULL;
  const char* place_name = place->getName();
  if (dup <= 0)
    place_dup_name = namecopy(place_name);
  else
  {
    int radix = 'z'-'a';
    dup--;
    if (dup > radix)
      Error("place <%s> has %d duplications while maximum allowed is %d", place_name, dup, radix);
    if ( strstr(&place_name[0], NOTATION_DUPLICATED_PLACE_SUFFIX) == NULL)
      place_dup_name = name2pattern("%s%s%c", place_name, NOTATION_DUPLICATED_PLACE_SUFFIX, 'a'+dup);
    else
      place_dup_name = name2pattern("%s%c", place_name, 'a'+dup);
  }
  return place_dup_name;
}

//---------------------------------------------------------------------------
unsigned GraphProtode::genTransitionDupIndex(Transition* transition, int dup)
{
  unsigned index = transition->getIndex();
  if (dup > 0)
  {
    if (dup >= TRANSITION_DUPLICATION_RADIX)
      Error("transition <%s> has %d duplications while maximum allowed is %d", 
        transition->getName(), dup, TRANSITION_DUPLICATION_RADIX);
    unsigned index_real = index % TRANSITION_REAL_RADIX;
    unsigned index_dup = index / TRANSITION_REAL_RADIX;
    index = (index_dup * TRANSITION_DUPLICATION_RADIX + dup) * TRANSITION_REAL_RADIX + index_real;
  }
  return index;
}

//---------------------------------------------------------------------------
Protocol* GraphProtode::addProtocol(const char* protocol_name)
{
  Protocol* protocol_new = new Protocol(protocol_name);
  Protocol* protocol = protocol_hashset->find(protocol_new);
  if (protocol)
    delete protocol_new;
  else
  {
    protocol = protocol_new;
    protocols->add(protocol);
    protocol_hashset->add(protocol);
  }
  return protocol;
}

//---------------------------------------------------------------------------
bool GraphProtode::addProtocol(Protocol* protocol)
{
  bool result = false;
  if ( !protocol_hashset->find(protocol) )
  {
    protocols->add(protocol);
    protocol_hashset->add(protocol);
    result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void GraphProtode::addProtocols(ProtocolArray* protocols)
{
  if (protocols)
  for (int i = 0; i < protocols->count(); i++)
  {
    Protocol* protocol_cur = (*protocols)[i];
    if ( !addProtocol(protocol_cur) )
      delete protocol_cur;
  }
}

//---------------------------------------------------------------------------
void GraphProtode::removeProtocol(Protocol* protocol)
{
  protocol_hashset->remove(protocol);
  protocols->remove(protocol, true);
}

//---------------------------------------------------------------------------
Protocol* GraphProtode::findProtocol(const char* protocol_name) const
{
  Protocol* protocol_mask = new Protocol(protocol_name);
  Protocol* protocol_cur = protocol_hashset->find(protocol_mask);
  delete protocol_mask;
  return protocol_cur;
}

//---------------------------------------------------------------------------
int GraphProtode::getProtocolObjectCount(const char* object_name) const
{
  int result = 0;
  for (int protocol_index = 0; protocol_index < getProtocolCount(); protocol_index++)
  {
    Protocol* protocol_cur = getProtocol(protocol_index);
    if (protocol_cur->getExtract() && protocol_cur->findObject(object_name))
      result++;
  }  
  return result;
}

//---------------------------------------------------------------------------
void GraphProtode::addProtocolsDevEnv()
{
  Protocol* protocol_dev = addProtocol(PROTOCOL_DEV_NAME);
  Protocol* protocol_env = addProtocol(PROTOCOL_ENV_NAME);
  for (int signal_index = 0; signal_index < signals->count(); signal_index++)
  {
    Signal* signal_cur = (*signals)[signal_index];
    if (signal_cur->getType() == Signal::TYPE_INPUT)
      protocol_env->addObject(signal_cur->getName());
    else
    if ( (signal_cur->getType() == Signal::TYPE_OUTPUT)
      || (signal_cur->getType() == Signal::TYPE_INTERNAL) )
      protocol_dev->addObject(signal_cur->getName());
  }
}

//---------------------------------------------------------------------------
void GraphProtode::addProtocolsSets(const char* signal_sets)
{
  if (signal_sets)
  {
    protocol_data = new ProtocolArray(2);
    protocol__scan_string(signal_sets);
    protocol_parse();
    addProtocols(protocol_data);
    delete protocol_data;
  }
}

//---------------------------------------------------------------------------
void GraphProtode::setProtocolsExtract(const char* protocol_names)
{
  if (protocol_names)
  {
    int name_index = 0;
    while (char* protocol_name = names2name(protocol_names, name_index++))
    {
      Protocol* protocol_cur = findProtocol(protocol_name);
      if (protocol_cur)
        protocol_cur->setExtract(true);
      delete [] protocol_name;
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::setProtocolsOptimise(const char* protocol_names, int optimisation_level)
{
  if (protocol_names)
  {
    int name_index = 0;
    while (char* protocol_name = names2name(protocol_names, name_index++))
    {
      Protocol* protocol_cur = findProtocol(protocol_name);
      if (protocol_cur)
        protocol_cur->setOptimisationLevel(optimisation_level);
      delete [] protocol_name;
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::setProtocolsWrite(const char* protocol_names)
{
  if (protocol_names)
  {
    int name_index = 0;
    while (char* protocol_name = names2name(protocol_names, name_index++))
    {
      Protocol* protocol_cur = findProtocol(protocol_name);
      if (protocol_cur)
        protocol_cur->setWrite(true);
      delete [] protocol_name;
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::calcProtocols()
{
  for (int protocol_index = 0; protocol_index < protocols->count(); protocol_index++)
  {
    Protocol* protocol_cur = (*protocols)[protocol_index];
    for (int object_index = 0; object_index < protocol_cur->getObjectCount(); object_index++)
    {
      Object* object_cur = protocol_cur->getObject(object_index);
      Signal* signal_cur = findSignal(object_cur->getName());
      if (signal_cur == NULL)
      {
        Warning("signal <%s> of protocol <%s> not found", object_cur->getName(), protocol_cur->getName());
        protocol_cur->removeObject(object_cur);
        object_index--;
      }
      else
      if ( (signal_cur->getType() != Signal::TYPE_INPUT)
        && (signal_cur->getType() != Signal::TYPE_OUTPUT)
        && (signal_cur->getType() != Signal::TYPE_INTERNAL) )
      {
        Warning("signal <%s> of protocol <%s> is not an interface signal", object_cur->getName(), protocol_cur->getName());
        protocol_cur->removeObject(object_cur);
        object_index--;
      }
    }
  }
}

//---------------------------------------------------------------------------
bool GraphProtode::checkProtocols()
{
  bool result = true;
  // prepare data
  for (int signal_index = 0; signal_index < signals->count(); signal_index++)
  {
    Signal* signal_cur = (*signals)[signal_index];
    signal_cur->setData( new int(0) );
  }
  // auxiliary caclulations
  for (int protocol_index = 0; protocol_index < protocols->count(); protocol_index++)
  {
    Protocol* protocol_cur = (*protocols)[protocol_index]; 
    if (protocol_cur->getExtract())
    {
      for (int object_index = 0; object_index < protocol_cur->getObjectCount(); object_index++)
      {
        Object* object_cur = protocol_cur->getObject(object_index);
        Signal* signal_cur = findSignal(object_cur->getName());
        if (signal_cur)
        {
          int* signal_data = (int*)signal_cur->getData();
          (*signal_data)++;
        }
      }
    }
  }
  // check, store and clean
  Protocol* protocol_misc = addProtocol(PROTOCOL_MISC_NAME);
  for (int signal_index = 0; signal_index < signals->count(); signal_index++)
  {
    Signal* signal_cur = (*signals)[signal_index];
    int* signal_data = (int*) signal_cur->getData();
    if ( (signal_cur->getType() == Signal::TYPE_INPUT)
      || (signal_cur->getType() == Signal::TYPE_OUTPUT)
      || (signal_cur->getType() == Signal::TYPE_INTERNAL) )
    {
      if ( *signal_data == 0)
        protocol_misc->addObject(signal_cur->getName());
      else
      if ( *signal_data >  1)
      {
        Warning("signal <%s> cannot be extracted into more than one protocol", signal_cur->getName());
        result = false;
      }
    }
    delete signal_data; signal_data = NULL;
  }
  if (protocol_misc->getObjectCount() > 0)
    protocol_misc->setExtract(true);
  return result;
}

//---------------------------------------------------------------------------
Graph* GraphProtode::addTracker(const char* tracker_name)
{
  Graph* tracker_new = new Graph(tracker_name);
  Graph* tracker = tracker_hashset->find(tracker_new);
  if (tracker)
    delete tracker_new;
  else
  {
    tracker = tracker_new;
    trackers->add(tracker);
    tracker_hashset->add(tracker);
  }
  return tracker;
}

//---------------------------------------------------------------------------
bool GraphProtode::addTracker(Graph* tracker)
{
  bool result = false;
  if ( !tracker_hashset->find(tracker) )
  {
    trackers->add(tracker);
    tracker_hashset->add(tracker);
    result = true;
  }
  return result;
}

//---------------------------------------------------------------------------
void GraphProtode::addTrackers(GraphArray* trackers)
{
  if (trackers)
  for (int i = 0; i < trackers->count(); i++)
  {
    Graph* tracker_cur = (*trackers)[i];
    if ( !addTracker(tracker_cur) )
      delete tracker_cur;
  }
}

//---------------------------------------------------------------------------
void GraphProtode::removeTracker(Graph* tracker)
{
  tracker_hashset->remove(tracker);
  trackers->remove(tracker, true);
}

//---------------------------------------------------------------------------
Graph* GraphProtode::findTracker(const char* tracker_name) const
{
  Graph* tracker_mask = new Graph(tracker_name);
  Graph* tracker_cur = tracker_hashset->find(tracker_mask);
  delete tracker_mask;
  return tracker_cur;
}

//---------------------------------------------------------------------------
void GraphProtode::exposeTransition(Graph* tracker, int transition_index, Place* place_from, Place* place_to)
{
  // create/find dummy signal
  Transition* transition_cur = tracker->getTransition(transition_index);
  char* signal_dummy_name = genSignalDummyName(transition_cur);
  Signal* signal_dummy = tracker->addSignal(signal_dummy_name, Signal::TYPE_DUMMY);
  delete [] signal_dummy_name;
  // create dummy transition whose data points to the original signal
  Signal* signal_cur = findSignal(transition_cur->getSignalName());
  Transition* transition_dummy = new Transition(signal_dummy->getName(), Transition::POLARITY_UNDEFINED, 
    transition_cur->getIndex(), TAG_GENERATED | TAG_TRANSITION_PROTOCOL_OUTPUT, signal_cur);
  tracker->addTransition(transition_dummy, transition_index);
  // expose rising transition of the signal
  swapTransitions(transition_cur, transition_dummy);
  Transition* transition_interface = new Transition(*transition_cur);
  interface->addTransition(transition_interface);
  tracker->removeTransition(transition_cur);
  addConsumingArc(place_from, transition_interface);
  addProducingArc(transition_interface, place_to);
  addReadArc(place_to, transition_dummy);
  addReadArcs(transition_dummy->getPreset(), transition_interface);
}

//---------------------------------------------------------------------------
void GraphProtode::exposeTransitions(Graph* tracker, Signal* signal, Place* place_low, Place* place_high)
{
  for (int transition_index = 0; transition_index < tracker->getTransitionCount(); transition_index++)
  {
    Transition* transition_cur = tracker->getTransition(transition_index);
    if ( strcmp(transition_cur->getSignalName(), signal->getName()) == 0 )
    {
      switch ( transition_cur->getPolarity() )
      {
        case Transition::POLARITY_RISING:
          exposeTransition(tracker, transition_index, place_low, place_high);
          break;
        case Transition::POLARITY_FALLING:
          exposeTransition(tracker, transition_index, place_high, place_low);
          break;
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::controlTransition(Graph* tracker, int transition_index, Place* place_to)
{
  // get dummy signal
  Transition* transition_cur = tracker->getTransition(transition_index);
  char* signal_dummy_name = genSignalDummyName(transition_cur);
  Signal* signal_dummy = tracker->addSignal(signal_dummy_name, Signal::TYPE_DUMMY);
  delete [] signal_dummy_name;
  // create dummy transition whose data points to the original signal
  Signal* signal_cur = findSignal(transition_cur->getSignalName());
  Transition* transition_dummy = new Transition(signal_dummy->getName(), Transition::POLARITY_UNDEFINED, 
    transition_cur->getIndex(), TAG_GENERATED | TAG_TRANSITION_PROTOCOL_INPUT, signal_cur);
  tracker->addTransition(transition_dummy, transition_index);
  // control dummy transition
  swapTransitions(transition_cur, transition_dummy);
  tracker->removeTransition(transition_cur);
  addReadArc(place_to, transition_dummy);
}

//---------------------------------------------------------------------------
void GraphProtode::controlTransitions(Graph* tracker, Signal* signal, Place* place_low, Place* place_high)
{
  for (int transition_index = 0; transition_index < tracker->getTransitionCount(); transition_index++)
  {
    Transition* transition_cur = tracker->getTransition(transition_index);
    if ( strcmp(transition_cur->getSignalName(), signal->getName()) == 0 )
    {
      switch ( transition_cur->getPolarity() )
      {
        case Transition::POLARITY_RISING:
          controlTransition(tracker, transition_index, place_high);
          break;
        case Transition::POLARITY_FALLING:
          controlTransition(tracker, transition_index, place_low);
          break;
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::exposeSignal(Signal* signal)
{
  // create places for low level of the signal
  char* place_low_name = name2pattern("%s%s", signal->getName(), NOTATION_EXPOSED_PLACE_SUFFIX_LOW);
  Place* place_low = new Place(place_low_name, 0, TAG_GENERATED);
  delete [] place_low_name;
  interface->addPlace(place_low);
  // create places for high level of the signal
  char* place_high_name = name2pattern("%s%s", signal->getName(), NOTATION_EXPOSED_PLACE_SUFFIX_HIGH);
  Place* place_high = new Place(place_high_name, 0, TAG_GENERATED);
  delete [] place_high_name;
  interface->addPlace(place_high);

  // set marking
  if (signal->getState() == Signal::STATE_LOW)
    place_low->setMarking(1);
  else 
    place_high->setMarking(1);
  // change all transitions of the signal to dummies
  Signal* signal_interface = new Signal(*signal);
  interface->addSignal(signal_interface);
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Signal* signal_cur = tracker_cur->findSignal(signal_interface->getName());
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if ( protocol_cur->findObject(signal_interface->getName()) )
      exposeTransitions(tracker_cur, signal_cur, place_low, place_high);
    else
      controlTransitions(tracker_cur, signal_cur, place_low, place_high);
    tracker_cur->removeSignal(signal_cur);
  }
}

//---------------------------------------------------------------------------
bool GraphProtode::hasArbitration(Place* place, Transition* transition)
{
  bool result = false;
  if (transition->getPostsetCount() == 1)
  {
    Signal* signal_cur = findSignal(transition->getSignalName());
    if ( (signal_cur->getType() == Signal::TYPE_INPUT)
      || (signal_cur->getType() == Signal::TYPE_OUTPUT) 
      || (signal_cur->getType() == Signal::TYPE_INTERNAL) )
    {
      Place* place_cur = transition->getPostset(0);
      for (int postset_index = 0; postset_index < place->getPostsetCount(); postset_index++)
      {
        Transition* transition_cur = place->getPostset(postset_index);
        if (isReadArc(place_cur, transition_cur))
        {
          result = true;
          break;
        }
      }
    }
  }
  return result;
}

//---------------------------------------------------------------------------
void GraphProtode::exposeArbitration(Place* place)
{
  for (int i = 0; i < place->getReadsetCount(); i++)
  {
    Transition* transition_1 = place->getReadset(i);
    if (hasArbitration(place, transition_1))
    {
      Place* place_1 = transition_1->getPreset(0);
      for (int j = i+1; j < place->getReadsetCount(); j++)
      {
        Transition* transition_2 = place->getReadset(j);
        if (hasArbitration(place, transition_2))
        {
          Place* place_2 = transition_2->getPreset(0);
          addReadArc(place_1, transition_2);
          addReadArc(place_2, transition_1);
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::extractProtocols()
{
  // create trackers
  for (int protocol_index = 0; protocol_index < getProtocolCount(); protocol_index++)
  {
    Protocol* protocol_cur = getProtocol(protocol_index); 
    if (protocol_cur->getExtract())
    {
      Graph* tracker = new Graph( *this );
      tracker->setName(protocol_cur->getName());
      addTracker(tracker);
    }
  }
  // expose interface
  for (int signal_index = 0; signal_index < getSignalCount(); signal_index++)
  {
    Signal* signal_cur = getSignal(signal_index);
    if (signal_cur->getType() != Signal::TYPE_DUMMY)
      exposeSignal(signal_cur);
  }
  // create mutual exclusion for exposed signals after choice
  for (int place_index = 0; place_index < getPlaceCount(); place_index++)
  {
    Place* place_cur = getPlace(place_index);
    if (place_cur->isChoice())
      exposeArbitration(place_cur);
  }
}

//---------------------------------------------------------------------------
void GraphProtode::findPlaceDeepPostset(Place* place, int depth, TransitionArray* transition_set, PlaceArray* place_set)
{
  if ( !place->testTag(TAG_PROCESSED) )
  {
    place->addTag(TAG_PROCESSED);
    if (place_set)
      place_set->add(place);
    if (depth > 0)
    {
      for (int transition_index = 0; transition_index < place->getPostsetCount(); transition_index++)
      {
        Transition* transition_cur = place->getPostset(transition_index);
        if ( !transition_cur->testTag(TAG_PROCESSED) )
        {
          transition_cur->addTag(TAG_PROCESSED);
          if (transition_set)
            transition_set->add(transition_cur);
          for (int place_index = 0; place_index < transition_cur->getPostsetCount(); place_index++)
          {
            Place* place_cur = transition_cur->getPostset(place_index);
            findPlaceDeepPostset(place_cur, depth-1, transition_set, place_set);
          }
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::findTransitionDeepPostset(Transition* transition, int depth, PlaceArray* place_set, TransitionArray* transition_set)
{
  if ( !transition->testTag(TAG_PROCESSED) )
  {
    transition->addTag(TAG_PROCESSED);
    if (transition_set)
      transition_set->add(transition);
    if (depth > 0)
    {
      for (int place_index = 0; place_index < transition->getPostsetCount(); place_index++)
      {
        Place* place_cur = transition->getPostset(place_index);
        if ( !place_cur->testTag(TAG_PROCESSED) )
        {
          place_cur->addTag(TAG_PROCESSED);
          if (place_set)
            place_set->add(place_cur);
          for (int transition_index = 0; transition_index < place_cur->getPostsetCount(); transition_index++)
          {
            Transition* transition_cur = place_cur->getPostset(transition_index);
            findTransitionDeepPostset(transition_cur, depth-1, place_set, transition_set);
          }
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::findPlaceDeepPreset(Place* place, int depth, TransitionArray* transition_set, PlaceArray* place_set)
{
  if ( !place->testTag(TAG_PROCESSED) )
  {
    place->addTag(TAG_PROCESSED);
    if (place_set)
      place_set->add(place);
    if (depth > 0)
    {
      for (int transition_index = 0; transition_index < place->getPresetCount(); transition_index++)
      {
        Transition* transition_cur = place->getPreset(transition_index);
        if ( !transition_cur->testTag(TAG_PROCESSED) )
        {
          transition_cur->addTag(TAG_PROCESSED);
          if (transition_set)
            transition_set->add(transition_cur);
          for (int place_index = 0; place_index < transition_cur->getPresetCount(); place_index++)
          {
            Place* place_cur = transition_cur->getPreset(place_index);
            findPlaceDeepPreset(place_cur, depth-1, transition_set, place_set);
          }
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::findTransitionDeepPreset(Transition* transition, int depth, PlaceArray* place_set, TransitionArray* transition_set)
{
  if ( !transition->testTag(TAG_PROCESSED) )
  {
    transition->addTag(TAG_PROCESSED);
    if (transition_set)
      transition_set->add(transition);
    if (depth > 0)
    {
      for (int place_index = 0; place_index < transition->getPresetCount(); place_index++)
      {
        Place* place_cur = transition->getPreset(place_index);
        if ( !place_cur->testTag(TAG_PROCESSED) )
        {
          place_cur->addTag(TAG_PROCESSED);
          if (place_set)
            place_set->add(place_cur);
          for (int transition_index = 0; transition_index < place_cur->getPresetCount(); transition_index++)
          {
            Transition* transition_cur = place_cur->getPreset(transition_index);
            findTransitionDeepPreset(transition_cur, depth-1, place_set, transition_set);
          }
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::findBackwardNeighbourhood(Place* place, TransitionArray* transition_set)
{
  if ( !place->testTag(TAG_PROCESSED) )
  {
    place->addTag(TAG_PROCESSED);
    for (int transition_index = 0; transition_index < place->getPresetCount(); transition_index++)
    {
      Transition* transition_cur = place->getPreset(transition_index);
      if ( !transition_cur->testTag(TAG_PROCESSED) )
      {
        transition_cur->addTag(TAG_PROCESSED);
        transition_set->add(transition_cur);
      }
      for (int place_index = 0; place_index < transition_cur->getPresetCount(); place_index++)
      {
        Place* place_cur = transition_cur->getPreset(place_index);
        if ( place_cur->testTag(TAG_PLACE_REDUNDANT) )
          findBackwardNeighbourhood(place_cur, transition_set);
      }
    }
  }
}

//---------------------------------------------------------------------------
TransitionArray* GraphProtode::calcBackwardNeighbourhood(Graph* tracker, Place* place)
{
  TransitionArray* transition_set = new TransitionArray(2);
  removeObjectsTag((ObjectArray*)tracker->getTransitions(), TAG_PROCESSED);
  removeObjectsTag((ObjectArray*)tracker->getPlaces(), TAG_PROCESSED);
  findBackwardNeighbourhood(place, transition_set);
  return transition_set;
}

//---------------------------------------------------------------------------
void GraphProtode::findForwardNeighbourhood(Place* place, TransitionArray* transition_set)
{
  if ( !place->testTag(TAG_PROCESSED) )
  {
    place->addTag(TAG_PROCESSED);
    for (int transition_index = 0; transition_index < place->getPostsetCount(); transition_index++)
    {
      Transition* transition_cur = place->getPostset(transition_index);
      if ( !transition_cur->testTag(TAG_PROCESSED) )
      {
        transition_cur->addTag(TAG_PROCESSED);
        transition_set->add(transition_cur);
      }
      for (int place_index = 0; place_index < transition_cur->getPostsetCount(); place_index++)
      {
        Place* place_cur = transition_cur->getPostset(place_index);
        if ( place_cur->testTag(TAG_PLACE_REDUNDANT) )
          findForwardNeighbourhood(place_cur, transition_set);
      }
    }
  }
}

//---------------------------------------------------------------------------
TransitionArray* GraphProtode::calcForwardNeighbourhood(Graph* tracker, Place* place)
{
  TransitionArray* transition_set = new TransitionArray(2);
  removeObjectsTag((ObjectArray*)tracker->getTransitions(), TAG_PROCESSED);
  removeObjectsTag((ObjectArray*)tracker->getPlaces(), TAG_PROCESSED);
  findForwardNeighbourhood(place, transition_set);
  return transition_set;
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocolChoice(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  for (int place_choice_index = 0; place_choice_index < tracker->getPlaceCount(); place_choice_index++)
  {
    Place* place_choice = tracker->getPlace(place_choice_index);
    // check if choice place postset is controlled by protocol inputs only
    if ( place_choice->isChoice() 
      && testObjectsTagAll((ObjectArray*)place_choice->getPostset(), TAG_TRANSITION_PROTOCOL_INPUT) )
    {
      for (int transition_index = 0; transition_index < place_choice->getPostsetCount(); transition_index++)
      {
        Transition* transition_cur = place_choice->getPostset(transition_index);
        // create a sets of transitions,  conflicting with the current choice branch
        removeObjectsTag((ObjectArray*)tracker->getSignals(), TAG_PROCESSED);
        removeObjectsTag((ObjectArray*)tracker->getTransitions(), TAG_PROCESSED);
        removeObjectsTag((ObjectArray*)tracker->getPlaces(), TAG_PROCESSED);
        TransitionArray* transitions_conf = new TransitionArray(3);
        findTransitionDeepPostset(transition_cur, 2, NULL, NULL);
        findPlaceDeepPostset(place_choice, 2, transitions_conf, NULL);
        // tag signals which control transitions in Tconf
        for (int transition_conf_index = 0; transition_conf_index < transitions_conf->count(); transition_conf_index++)
        {
          Transition* transition_conf = (*transitions_conf)[transition_conf_index];
          Signal* signal_conf = (Signal*)transition_conf->getData();
          signal_conf->addTag(TAG_PROCESSED);
        }
        // check signals in the current choice branch
        for (int place_post_index = 0; place_post_index < transition_cur->getPostsetCount(); place_post_index++)
        {
          Place* place_post = transition_cur->getPostset(place_post_index);
          PlaceArray* places_post_post = new PlaceArray(3);
          bool state_machine_conflict = false;
          for (int transition_post_index = 0; transition_post_index < place_post->getPostsetCount(); transition_post_index++)
          {
            Transition* transition_post = place_post->getPostset(transition_post_index);
            Signal* signal_post = (Signal*)transition_post->getData();
            if (signal_post->testTag(TAG_PROCESSED))
            {
              place_post->addTag(TAG_PLACE_MANDATORY);
              state_machine_conflict = true;
              break;
            }
            places_post_post->merge( *(transition_post->getPostset()) );
          }
          if ( !state_machine_conflict )
          {
            place_post->addTag(TAG_PLACE_REDUNDANT);
            for (int place_index = 0; place_index < places_post_post->count(); place_index++)
            {
              Place* place_cur = (*places_post_post)[place_index];
              place_cur->addTag(TAG_PLACE_MANDATORY);
            }
          }
          delete places_post_post;
        }
        delete transitions_conf;
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocolLatency(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
  {
    Place* place_cur = tracker->getPlace(place_index);
    if ( !place_cur->testTag(TAG_PLACE_MANDATORY)
      && !place_cur->testTag(TAG_PLACE_REDUNDANT) )
    {
      if ( testObjectsTagAll((ObjectArray*)place_cur->getPreset(), TAG_TRANSITION_PROTOCOL_INPUT) 
        && testObjectsTagAll((ObjectArray*)place_cur->getPostset(), TAG_TRANSITION_PROTOCOL_OUTPUT) )
        place_cur->addTag(TAG_PLACE_REDUNDANT);
    }
  }
}

//---------------------------------------------------------------------------
// check the intersection of signals in backward and foward neighbouhoods
bool GraphProtode::isTrackerPlaceRedundant(Graph* tracker, Place* place)
{
  bool marked_graph_conflict = false;
  removeObjectsTag((ObjectArray*)tracker->getSignals(), TAG_PROCESSED);
  TransitionArray* neighbourhooh_backward = calcBackwardNeighbourhood(tracker, place);
  for (int transition_index = 0; transition_index < neighbourhooh_backward->count(); transition_index++)
  {
    Transition* transition_cur = (*neighbourhooh_backward)[transition_index];
    Signal* signal_cur = (Signal*) transition_cur->getData();
    if (signal_cur)
      signal_cur->addTag(TAG_PROCESSED);
  }
  delete neighbourhooh_backward;
  TransitionArray* neighbourhooh_forward = calcForwardNeighbourhood(tracker, place);
  for (int transition_index = 0; transition_index < neighbourhooh_forward->count(); transition_index++)
  {
    Transition* transition_cur = (*neighbourhooh_forward)[transition_index];
    Signal* signal_cur = (Signal*) transition_cur->getData();
    if (signal_cur && signal_cur->testTag(TAG_PROCESSED))
    {
      marked_graph_conflict = true;
      break;
    }
  }
  delete neighbourhooh_forward;
  return (!marked_graph_conflict);
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocolSizeChain(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  bool modified = true;
  while (modified)
  {
    modified = false;
    for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
    {
      Place* place_cur = tracker->getPlace(place_index);
      if ( !place_cur->testTag(TAG_PLACE_MANDATORY)
        && !place_cur->testTag(TAG_PLACE_REDUNDANT) )
      {
        bool border = true;
        // test pre-preset
        if (border)
        {
          for (int transition_pred_index = 0; transition_pred_index < place_cur->getPresetCount(); transition_pred_index++)
          {
            Transition* transition_pred = place_cur->getPreset(transition_pred_index);
            border = inObjectsTagAll((ObjectArray*)transition_pred->getPreset(), TAG_PLACE_REDUNDANT | TAG_PLACE_MANDATORY);
          }
        }
        // test post-postset
        if (border)
        {
          for (int transition_succ_index = 0; transition_succ_index < place_cur->getPostsetCount(); transition_succ_index++)
          {
            Transition* transition_succ = place_cur->getPostset(transition_succ_index);
            border = !inObjectsTagAny((ObjectArray*)transition_succ->getPreset(), TAG_PLACE_REDUNDANT | TAG_PLACE_MANDATORY);
          }
        }
        // test redundancy
        if (border)
        {
          modified = true;
          if ( isTrackerPlaceRedundant(tracker, place_cur) )
            place_cur->addTag(TAG_PLACE_REDUNDANT);
          else
            place_cur->addTag(TAG_PLACE_MANDATORY);
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocolSizeIndividual(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
  {
    Place* place_cur = tracker->getPlace(place_index);
    if ( !place_cur->testTag(TAG_PLACE_MANDATORY)
      && !place_cur->testTag(TAG_PLACE_REDUNDANT) )
    {
      if ( isTrackerPlaceRedundant(tracker, place_cur) )
        place_cur->addTag(TAG_PLACE_REDUNDANT);
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocolMandatary(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
  {
    Place* place_cur = tracker->getPlace(place_index);
    if ( !place_cur->testTag(TAG_PLACE_MANDATORY)
      && !place_cur->testTag(TAG_PLACE_REDUNDANT) )
    {
      place_cur->addTag(TAG_PLACE_MANDATORY);
    }
  }
}

//---------------------------------------------------------------------------
// Find the place preset transition whose all postset places are marked and 
// the differnce between the number of its postset places and preset places 
// is maximal.
Transition* GraphProtode::findBestRollbackTransition(Place* place)
{
  Transition* transition_best = NULL;
  int difference_max = -1;
  for (int transition_index = 0; transition_index < place->getPresetCount(); transition_index++)
  {
    Transition* transition_cur = place->getPreset(transition_index);
    bool marked_postset = true;
    for (int place_index = 0; place_index < transition_cur->getPostsetCount(); place_index++)
    {
      Place* place_post = transition_cur->getPostset(place_index);
      if (place_post->getMarking() == 0)
      {
        marked_postset = false;
        break;
      }
    }
    if (marked_postset)
    {
      int difference = transition_cur->getPostsetCount() - transition_cur->getPresetCount();
      if (difference > difference_max)
      {
        transition_best = transition_cur;
        difference_max = difference;
      }
    }
  } 
  return transition_best;
}

//---------------------------------------------------------------------------
void GraphProtode::transformProtocolMarking(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  // try to roll back the marking from redundant places
  bool done = false;
  while ( !done )
  {
    done = true;
    for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
    {
      Place* place_cur = tracker->getPlace(place_index);
      if (place_cur->testTag(TAG_PLACE_REDUNDANT) && (place_cur->getMarking() > 0))
      {
        Transition* transition_best = findBestRollbackTransition(place_cur);
        if (transition_best)
        {
          for (int place_post_index = 0; place_post_index < transition_best->getPostsetCount(); place_post_index++)
          {
            Place* place_post = transition_best->getPostset(place_post_index);
            place_post->decMarking();
          }
          for (int place_pre_index = 0; place_pre_index < transition_best->getPresetCount(); place_pre_index++)
          {
            Place* place_pre = transition_best->getPreset(place_pre_index);
            place_pre->incMarking();
          }
          done = false;
        }
      }
    }
  }
  // if rolling back the marking for some redundant places is impossible, then make these places mandatory
  for (int place_index = 0; place_index < tracker->getPlaceCount(); place_index++)
  {
    Place* place_cur = tracker->getPlace(place_index);
    if (place_cur->testTag(TAG_PLACE_REDUNDANT) && (place_cur->getMarking() > 0))
    {
      place_cur->removeTag(TAG_PLACE_REDUNDANT);
      place_cur->addTag(TAG_PLACE_MANDATORY);
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::transformBouncerReadArc(Graph* tracker, Place* place_redundant, Transition* transition_read)
{
  int dup = 0;
  int transition_count = place_redundant->getPresetCount();
  TransitionArray* transitions_read_dup = new TransitionArray(transition_count);
  for (int transition_index = transition_count-1; transition_index >= 0; transition_index--)
  {
    Transition* transition_cur = place_redundant->getPreset(transition_index);
    Message("  locate transition_cur <%s> preceeding redundant place_redundant <%s>", transition_cur->getName(), place_redundant->getName());
    // dup transition_read
    Transition* transition_read_dup = new Transition(transition_read->getName(), transition_read->getTag(), transition_read->getData());
    if (transition_count > 1)
      transition_read_dup->setIndex( genTransitionDupIndex(transition_read, transition_index+1) );
    transitions_read_dup->add(transition_read_dup);
    Message("    duplicate transition_read <%s> into transition_read_dup <%s>", transition_read->getName(), transition_read_dup->getName());
    addConsumingArcs(transition_read->getPreset(), transition_read_dup);
    addProducingArcs(transition_read_dup, transition_read->getPostset());
    addReadArcs(transition_read->getReadset(), transition_read_dup, place_redundant);
    // form triggers (if it has not been formed yet)
    if ( transition_read->testTag(TAG_PROCESSED) )
      Message("      transition_read <%s> has trigers already", transition_read_dup->getName());
    else
    {
      Message("      build triggers for transition_read <%s>", transition_read_dup->getName());
      addReadArcs(transition_cur->getReadset(), transition_read_dup);
      transition_read_dup->addTag(TAG_PROCESSED);
    }
    // form context
    Message("      build context for interface transition <%s>", transition_read_dup->getName());
    Message("        duplicate transition_cur <%s> and it incident arcs", transition_cur->getName());
    Transition* transition_dup = new Transition(transition_cur->getName(), transition_cur->getTag(), transition_cur->getData());
    addProducingArc(transition_dup, place_redundant);
    addReadArcs(transition_cur->getReadset(), transition_dup);
    removeProducingArc(transition_cur, place_redundant);
    PlaceArray* places_pred_dup = new PlaceArray(transition_cur->getPresetCount());
    for (int place_pred_index = 0; place_pred_index < transition_cur->getPresetCount(); place_pred_index++)
    {
      Place* place_pred = transition_cur->getPreset(place_pred_index);
      Message("          duplicate place_pred <%s> and its incedent arcs", place_pred->getName());
      Place* place_pred_dup = new Place(place_pred->getName(), place_pred->getMarking(), place_pred->getTag(), place_pred->getData());
      places_pred_dup->add(place_pred_dup);
      addReadArc(place_pred_dup, transition_read_dup);
      addConsumingArcs(place_pred_dup, place_pred->getPostset(), transition_cur);
      addProducingArcs(place_pred->getPreset(), place_pred_dup);
      addReadArcs(place_pred_dup, place_pred->getReadset());
      addConsumingArc(place_pred_dup, transition_dup);
    }
    if (transition_cur->getPostsetCount() > 0)
    {
      Message("        give unique names to transition_dup <%s> and its preset", transition_dup->getName());
      dup++;
      for (int place_pred_index = 0; place_pred_index < places_pred_dup->count(); place_pred_index++)
      {
        Place* place_pred = (*places_pred_dup)[place_pred_index];
        char* place_pred_name = genPlaceDupName(place_pred, dup);
        place_pred->setName(place_pred_name);
        delete [] place_pred_name;
      }
      transition_dup->setIndex( genTransitionDupIndex(transition_cur, dup) );
    }
    else
    {
      Message("        remove transition_cur <%s> and its preset", transition_cur->getName());
      tracker->removePlaces(transition_cur->getPreset());
      tracker->removeTransition(transition_cur);
    }
    Message("        add transition_dup <%s> and its preset", transition_dup->getName());
    tracker->addTransition(transition_dup);
    tracker->addPlaces(places_pred_dup);
    delete places_pred_dup;
  }
  Message("    remove transition_read <%s>", transition_read->getName());
  interface->removeTransition(transition_read);
  interface->addTransitions(transitions_read_dup);
  delete transitions_read_dup;
}

//---------------------------------------------------------------------------
void GraphProtode::transformProtocolBouncer(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  Progress("Transform protocol <%s> bouncer", protocol->getName());
  removeObjectsTag((ObjectArray*)interface->getTransitions(), TAG_PROCESSED);
  bool modified = true;
  while (modified)
  {
    modified = false;
    for (int place_redundant_index = 0; place_redundant_index < tracker->getPlaceCount(); place_redundant_index++)
    {
      Place* place_redundant = tracker->getPlace(place_redundant_index);
      if (place_redundant->testTag(TAG_PLACE_REDUNDANT))
      {
        for (int transition_read_index = 0; transition_read_index < place_redundant->getReadsetCount(); transition_read_index++)
        {
          Transition* transition_read = place_redundant->getReadset(transition_read_index);
          Message("locate transition_read <%s> (%d) controlled by redundant place_redundant <%s> (%d)", 
            transition_read->getName(), transition_read_index, place_redundant->getName(), place_redundant_index);
          transformBouncerReadArc(tracker, place_redundant, transition_read);
          modified = true;
        }
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::transformTrackerRedundantPlace(Graph* tracker, Place* place_redundant)
{
  int transition_count = place_redundant->getPresetCount();
  for (int transition_index = transition_count-1; transition_index >=0 ; transition_index--)
  {
    Transition* transition_cur = place_redundant->getPreset(transition_index);
    Message("  locate transition_cur <%s> preceeding redundant place_redundant <%s>", transition_cur->getName(), place_redundant->getName());
    // duplicate transitions in the postset of place_redundant and their incident arcs
    TransitionArray* transitions_succ_dup = new TransitionArray(place_redundant->getPostsetCount());
    for (int transition_succ_index = 0; transition_succ_index < place_redundant->getPostsetCount(); transition_succ_index++)
    {
      Transition* transition_succ = place_redundant->getPostset(transition_succ_index);
      Transition* transition_succ_dup = new Transition(transition_succ->getName(), transition_succ->getTag(), transition_succ->getData());
      if (transition_count > 1)
        transition_succ_dup->setIndex( genTransitionDupIndex(transition_succ, transition_index+1) );
      transitions_succ_dup->add(transition_succ_dup);
      Message("    duplicate transition_succ <%s> into transition_succ_dup <%s>", transition_succ->getName(), transition_succ_dup->getName());
      addConsumingArcs(transition_succ->getPreset(), transition_succ_dup, place_redundant);
      addProducingArcs(transition_succ_dup, transition_succ->getPostset());
      addReadArcs(transition_succ->getReadset(), transition_succ_dup);
    }
    // connect transition_cur preset places with each transitions_succ_dup transition
    for (int place_pred_index = 0; place_pred_index < transition_cur->getPresetCount(); place_pred_index++)
    {
      Place* place_pred = transition_cur->getPreset(place_pred_index);
      addConsumingArcs(place_pred, transitions_succ_dup);
    }
    Message("  remove producing arcs from transition <%s> to place <%s>", transition_cur->getName(), place_redundant->getName());
    removeProducingArc(transition_cur, place_redundant);
    if (transition_cur->getPostsetCount() == 0)
    {
      Message("    remove transition <%s>", transition_cur->getName());
      tracker->removeTransition(transition_cur);
    }
    if (transition_index == 0)
    {
      Message("    remove place_redundant <%s> and its postset transitions", place_redundant->getName());
      tracker->removeTransitions(place_redundant->getPostset());
      tracker->removePlace(place_redundant);
    }
    tracker->addTransitions(transitions_succ_dup);
    delete transitions_succ_dup;
  }
}

//---------------------------------------------------------------------------
void GraphProtode::transformProtocolTracker(Protocol* protocol)
{
  Graph* tracker = findTracker(protocol->getName());
  if (tracker == NULL) return;
  Progress("Transform protocol <%s> tracker", protocol->getName());
  bool modified = true;
  while (modified)
  {
    modified = false;
    for (int place_redundant_index = 0; place_redundant_index < tracker->getPlaceCount(); place_redundant_index++)
    {
      Place* place_redundant = tracker->getPlace(place_redundant_index);
      if (place_redundant->testTag(TAG_PLACE_REDUNDANT))
      {
        Message("locate redundant place_redundant <%s> (%d)", place_redundant->getName(), place_redundant_index);
        transformTrackerRedundantPlace(tracker, place_redundant);
        modified = true;
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::optimiseProtocols()
{
  for (int protocol_index = 0; protocol_index < getProtocolCount(); protocol_index++)
  {
    Protocol* protocol_cur = getProtocol(protocol_index); 
    int optimisation_level = protocol_cur->getOptimisationLevel();
    if (optimisation_level > 0)
    {

      optimiseProtocolChoice(protocol_cur);
      if (optimisation_level > 1)
      {
        optimiseProtocolLatency(protocol_cur);
        if (optimisation_level > 2)
        {
          optimiseProtocolSizeChain(protocol_cur);
          optimiseProtocolSizeIndividual(protocol_cur);
        }
      }
    }
    optimiseProtocolMandatary(protocol_cur);
    // remove redundant places
    transformProtocolMarking(protocol_cur);
    transformProtocolBouncer(protocol_cur);
    transformProtocolTracker(protocol_cur);
  }
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgSignals(FILE* file, Signal::Type type, const char* label)
{
  bool label_printed = false;
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int signal_index = 0; signal_index < tracker_cur->getSignalCount(); signal_index++)
      {
        Signal* signal_cur = tracker_cur->getSignal(signal_index);
        if (signal_cur->getType() == type)
        {
          if (!label_printed)
          {
            fprintf(file, label);
            label_printed = true;
          }
          fprintf(file, " %s%s%s%s", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, signal_cur->getName());
        }
      }
    }
  }
  // interface
  for (int signal_index = 0; signal_index < interface->getSignalCount(); signal_index++)
  {
    Signal* signal_cur = interface->getSignal(signal_index);
    if (signal_cur->getType() == type)
    {
      if (!label_printed)
      {
        fprintf(file, label);
        label_printed = true;
      }
      fprintf(file, " %s", signal_cur->getName());
    }
  }
  if (label_printed)
    fprintf(file, "\n");
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgInitialStates(FILE* file)
{
  bool label_printed = false;
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int signal_index = 0; signal_index < tracker_cur->getSignalCount(); signal_index++)
      {
        Signal* signal_cur = tracker_cur->getSignal(signal_index);
        if ( (signal_cur->getState() == Signal::STATE_LOW)
          || (signal_cur->getState() == Signal::STATE_HIGH) )
        {
          if (!label_printed)
          {
            fprintf(file, ASTG_KEYWORD_INITIAL_STATE);
            label_printed = true;
          }
          fprintf(file, " ");
          if (signal_cur->getState() == Signal::STATE_LOW)
            fprintf( file, ASTG_PROTOCOL_PATTERN_LOW, tracker_cur->getName(), signal_cur->getName() );
          if (signal_cur->getState() == Signal::STATE_HIGH)
            fprintf( file, ASTG_PROTOCOL_PATTERN_HIGH, tracker_cur->getName(), signal_cur->getName() );
        }
      }
    }
  }
  // interface
  for (int signal_index = 0; signal_index < interface->getSignalCount(); signal_index++)
  {
    Signal* signal_cur = interface->getSignal(signal_index);
    if ( (signal_cur->getState() == Signal::STATE_LOW)
      || (signal_cur->getState() == Signal::STATE_HIGH) )
    {
      if (!label_printed)
      {
        fprintf(file, ASTG_KEYWORD_INITIAL_STATE);
        label_printed = true;
      }
      fprintf(file, " ");
      if (signal_cur->getState() == Signal::STATE_LOW)
        fprintf( file, ASTG_PATTERN_LOW, signal_cur->getName() );
      if (signal_cur->getState() == Signal::STATE_HIGH)
        fprintf( file, ASTG_PATTERN_HIGH, signal_cur->getName() );
    }
  }
  if (label_printed)
    fprintf(file, "\n");
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgProducingArcs(FILE* file, bool show_implicit)
{
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int transition_index = 0; transition_index < tracker_cur->getTransitionCount(); transition_index++)
      {
        Transition* transition_cur = tracker_cur->getTransition(transition_index);
        if (transition_cur->getPostsetCount() > 0)
        {
          fprintf(file, "%s%s%s%s", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, transition_cur->getName());
          for (int place_index = 0; place_index < transition_cur->getPostsetCount(); place_index++)
          {
            Place* place_cur = transition_cur->getPostset(place_index);
            fprintf(file, " ");
            if ( show_implicit || !place_cur->isImplicit() )
            {
              if (place_cur->testTag(TAG_PLACE_REDUNDANT))
                fprintf(file, NOTATION_REDUNDANT_PLACE_PREFIX);
              fprintf(file, ASTG_PROTOCOL_PATTERN_PLACE, tracker_cur->getName(), place_cur->getName() );
            }
            else  
              fprintf(file, "%s%s%s%s", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, place_cur->getPostset(0)->getName());
          }
          fprintf(file, "\n");
        }
      }
    }
  }
  // interface
  for (int transition_index = 0; transition_index < interface->getTransitionCount(); transition_index++)
  {
    Transition* transition_cur = interface->getTransition(transition_index);
    if (transition_cur->getPostsetCount() > 0)
    {
      fprintf(file, "%s", transition_cur->getName());
      for (int place_index = 0; place_index < transition_cur->getPostsetCount(); place_index++)
      {
        Place* place_cur = transition_cur->getPostset(place_index);
        if ( show_implicit || !place_cur->isImplicit() )
          fprintf(file, " %s", place_cur->getName() );
        else  
          fprintf(file, " %s", place_cur->getPostset(0)->getName());
      }
      fprintf(file, "\n");
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgConsumingArcs(FILE* file, bool show_implicit)
{
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int place_index = 0; place_index < tracker_cur->getPlaceCount(); place_index++)
      {
        Place* place_cur = tracker_cur->getPlace(place_index);
        if (place_cur->getPostsetCount() > 0)
        {
          if ( show_implicit || !place_cur->isImplicit() )
          {
            if (place_cur->testTag(TAG_PLACE_REDUNDANT))
              fprintf(file, NOTATION_REDUNDANT_PLACE_PREFIX);
            fprintf(file, ASTG_PROTOCOL_PATTERN_PLACE, tracker_cur->getName(), place_cur->getName());
            for (int transition_index = 0; transition_index < place_cur->getPostsetCount(); transition_index++)
            {
              Transition* transition_cur = place_cur->getPostset(transition_index);
              fprintf(file, " %s%s%s%s", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, transition_cur->getName());
            }
            fprintf(file, "\n");
          }
        }
      }
    }
  }
  // interface
  for (int place_index = 0; place_index < interface->getPlaceCount(); place_index++)
  {
    Place* place_cur = interface->getPlace(place_index);
    if (place_cur->getPostsetCount() > 0)
    {
      if ( show_implicit || !place_cur->isImplicit() )
      {
        fprintf(file, "%s", place_cur->getName());
        for (int transition_index = 0; transition_index < place_cur->getPostsetCount(); transition_index++)
        {
          Transition* transition_cur = place_cur->getPostset(transition_index);
          fprintf(file, " %s", transition_cur->getName());
        }
        fprintf(file, "\n");
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgReadArcs(FILE* file)
{
  // prepare data
  addObjectsTag((ObjectArray*)interface->getTransitions(), TAG_PROCESSED);
  addObjectsTag((ObjectArray*)interface->getPlaces(), TAG_PROCESSED);
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    removeObjectsTag((ObjectArray*)tracker_cur->getTransitions(), TAG_PROCESSED);
    removeObjectsTag((ObjectArray*)tracker_cur->getPlaces(), TAG_PROCESSED);
  }
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int place_index = 0; place_index < tracker_cur->getPlaceCount(); place_index++)
      {
        Place* place_cur = tracker_cur->getPlace(place_index);
        for (int transition_index = 0; transition_index < place_cur->getReadsetCount(); transition_index++)
        {
          Transition* transition_cur = place_cur->getReadset(transition_index);
          if (place_cur->testTag(TAG_PLACE_REDUNDANT))
            fprintf(file, NOTATION_REDUNDANT_PLACE_PREFIX);
          fprintf(file, ASTG_PROTOCOL_PATTERN_PLACE, tracker_cur->getName(), place_cur->getName());
          fprintf(file, " %s\n", transition_cur->getName());
          fprintf(file, "%s ", transition_cur->getName());
          if (place_cur->testTag(TAG_PLACE_REDUNDANT))
            fprintf(file, NOTATION_REDUNDANT_PLACE_PREFIX);
          fprintf(file, ASTG_PROTOCOL_PATTERN_PLACE, tracker_cur->getName(), place_cur->getName());
          fprintf(file, "\n");
        }
      }
      for (int transition_index = 0; transition_index < tracker_cur->getTransitionCount(); transition_index++)
      {
        Transition* transition_cur = tracker_cur->getTransition(transition_index);
        for (int place_index = 0; place_index < transition_cur->getReadsetCount(); place_index++)
        {
          Place* place_cur = transition_cur->getReadset(place_index);
          fprintf(file, "%s%s%s%s", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, transition_cur->getName());
          fprintf(file, " %s\n", place_cur->getName());
          fprintf(file, "%s ", place_cur->getName());
          fprintf(file, "%s%s%s%s\n", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, transition_cur->getName());
        }
      }
    }
  }
  // interface
  for (int place_index = 0; place_index < interface->getPlaceCount(); place_index++)
  {
    Place* place_cur = interface->getPlace(place_index);
    for (int transition_index = 0; transition_index < place_cur->getReadsetCount(); transition_index++)
    {
      Transition* transition_cur = place_cur->getReadset(transition_index);
      if (transition_cur->testTag(TAG_PROCESSED))
      {
        fprintf(file, "%s %s\n", place_cur->getName(), transition_cur->getName());
        fprintf(file, "%s %s\n", transition_cur->getName(), place_cur->getName());
      }
    }
  }
  for (int transition_index = 0; transition_index < interface->getTransitionCount(); transition_index++)
  {
    Transition* transition_cur = interface->getTransition(transition_index);
    for (int place_index = 0; place_index < transition_cur->getReadsetCount(); place_index++)
    {
      Place* place_cur = transition_cur->getReadset(place_index);
      if (place_cur->testTag(TAG_PROCESSED))
      {
        fprintf(file, "%s %s\n", transition_cur->getName(), place_cur->getName());
        fprintf(file, "%s %s\n", place_cur->getName(), transition_cur->getName());
      }
    }
  }
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgArcs(FILE* file, bool show_implicit)
{
  fprintf(file, "%s\n", ASTG_KEYWORD_GRAPH);
  writeProtocolsAstgProducingArcs(file, show_implicit);
  writeProtocolsAstgConsumingArcs(file, show_implicit);
  writeProtocolsAstgReadArcs(file);
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstgMarking(FILE* file, bool show_implicit)
{
  fprintf(file, "%s {", ASTG_KEYWORD_MARKING);
  bool label_printed = false;
  // trackers
  for (int tracker_index = 0; tracker_index < getTrackerCount(); tracker_index++)
  {
    Graph* tracker_cur = getTracker(tracker_index);
    Protocol* protocol_cur = findProtocol(tracker_cur->getName());
    if (protocol_cur && (protocol_cur->getWrite()) )
    {
      for (int place_index = 0; place_index < tracker_cur->getPlaceCount(); place_index++)
      {
        Place* place_cur = tracker_cur->getPlace(place_index);
        if (place_cur->getMarking() > 0)
        {
          if (label_printed)
            fprintf(file, " ");
          if ( show_implicit || !place_cur->isImplicit() )
          {
            if (place_cur->testTag(TAG_PLACE_REDUNDANT))
              fprintf(file, NOTATION_REDUNDANT_PLACE_PREFIX);
            fprintf(file, ASTG_PROTOCOL_PATTERN_PLACE, tracker_cur->getName(), place_cur->getName());
          }
          else
          {
            fprintf(file, "<%s%s%s%s,", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, place_cur->getPreset(0)->getName());
            fprintf(file, "%s%s%s%s>", NOTATION_PROTOCOL_DUMMY_PREFIX, tracker_cur->getName(), NOTATION_PROTOCOL_DUMMY_SUFFIX, place_cur->getPostset(0)->getName());
          }
          label_printed = true;
        }
      }
    }
  }
  // interface
  for (int place_index = 0; place_index < interface->getPlaceCount(); place_index++)
  {
    Place* place_cur = interface->getPlace(place_index);
    if (place_cur->getMarking() > 0)
    {
      if (label_printed)
        fprintf(file, " ");
      if ( show_implicit || !place_cur->isImplicit() )
        fprintf(file, "%s", place_cur->getName());
      else
        fprintf(file, "<%s, %s>", place_cur->getPreset(0)->getName(), place_cur->getPostset(0)->getName());
      label_printed = true;
    }
  }
  fprintf(file, "}\n");
}

//---------------------------------------------------------------------------
void GraphProtode::writeProtocolsAstg(const char* file_name, bool show_implicit)
{
  if (file_name)
  {
    FILE* file = NULL;
    if (strcmp(file_name, NAME_STDIO) == 0)
      file = stdout;
    else
    {
      if ((file = fopen(file_name, "w")) == NULL)
        Error("opening output file <%s>", file_name);
    }
    // model
    if (model)
      fprintf(file, "%s %s\n", ASTG_KEYWORD_MODEL, model);
    // signals
    writeProtocolsAstgSignals(file, Signal::TYPE_INPUT, ASTG_KEYWORD_INPUTS);
    writeProtocolsAstgSignals(file, Signal::TYPE_OUTPUT, ASTG_KEYWORD_OUTPUTS);
    writeProtocolsAstgSignals(file, Signal::TYPE_INTERNAL, ASTG_KEYWORD_INTERNAL);
    writeProtocolsAstgSignals(file, Signal::TYPE_DUMMY, ASTG_KEYWORD_DUMMY);
    // initial states
    writeProtocolsAstgInitialStates(file);
    // graph
    writeProtocolsAstgArcs(file, show_implicit);
    // marking
    writeProtocolsAstgMarking(file, show_implicit);
    // end
    fprintf(file, "%s\n", ASTG_KEYWORD_END);
    if (file != stdout)
      fclose(file);
  }
}
