#ifndef _STG_TRANSITION_H_
#define _STG_TRANSITION_H_

#include "array.h"
#include "hashset.h"
#include "stg_object.h"

class Transition;
typedef Array<Transition, ObjectCompareName> TransitionArray;
typedef HashSet<Transition, ObjectHashName, ObjectEqualName> TransitionHashSet;

#include "stg_place.h"

extern const char* TRANSITION_PATTERN_DUMMY;
extern const char* TRANSITION_PATTERN_SIMPLE;
extern const char* TRANSITION_PATTERN_INDEX;

extern const char TRANSITION_SYMBOL_RISING;
extern const char TRANSITION_SYMBOL_FALLING;
extern const char TRANSITION_SYMBOL_INDEX;


//---------------------------------------------------------------------------
class Transition : public Object
{
  public:
    enum Polarity 
    { 
      POLARITY_UNDEFINED, 
      POLARITY_RISING, 
      POLARITY_FALLING 
    };

    Transition(const char* _name, Tag _tag = TAG_UNDEFINED, void* _data = NULL);
    Transition(const char* _signal_name, Polarity _polarity, unsigned _index, Tag _tag = TAG_UNDEFINED, void* _data = NULL);
    Transition(const Transition& transition);
    ~Transition();

    void setName(const char* _name);

    void setSignalName(const char* _signal_name);
    const char* getSignalName() const;

    void setPolarity(Polarity _polarity);
    Polarity getPolarity() const;

    void setIndex(unsigned _index);
    unsigned getIndex() const;
    
    bool isFork() const;
    bool isJoin() const;

    int getPresetCount() const;
    void setPreset(PlaceArray* _preset);
    PlaceArray* getPreset() const;
    Place* getPreset(int position) const;
    void addPreset(Place* place);
    void replacePreset(Place* place, int position);
    void removePreset(Place* place);
    int locatePreset(Place* place) const;

    int getPostsetCount() const;
    void setPostset(PlaceArray* _postset);
    PlaceArray* getPostset() const;
    Place* getPostset(int position) const;
    void addPostset(Place* place);
    void replacePostset(Place* place, int position);
    void removePostset(Place* place);
    int locatePostset(Place* place) const;

    int getReadsetCount() const;
    void setReadset(PlaceArray* _readset);
    PlaceArray* getReadset() const;
    Place* getReadset(int position) const;
    void addReadset(Place* place);
    void replaceReadset(Place* place, int position);
    void removeReadset(Place* place);
    int locateReadset(Place* place) const;
    
  protected:
    char* signal_name;
    Polarity polarity;
    unsigned index;
    PlaceArray* preset;
    PlaceArray* postset;
    PlaceArray* readset;

    void genName();
    void parseName();
};

//---------------------------------------------------------------------------
inline void Transition::setName(const char* _name)
{
  Object::setName(_name);
  parseName();
}

//---------------------------------------------------------------------------
inline void Transition::setSignalName(const char* _signal_name)
{
  delete [] signal_name;
  signal_name = namecopy(_signal_name);
  genName();
}

//---------------------------------------------------------------------------
inline const char* Transition::getSignalName() const
{
  return signal_name;
}

//---------------------------------------------------------------------------
inline void Transition::setPolarity(Transition::Polarity _polarity)
{
  polarity = _polarity;
  genName();
}

//---------------------------------------------------------------------------
inline Transition::Polarity Transition::getPolarity() const
{
  return polarity;
}

//---------------------------------------------------------------------------
inline void Transition::setIndex(unsigned _index)
{
  index = _index;
  genName();
}

//---------------------------------------------------------------------------
inline unsigned Transition::getIndex() const
{
  return index;
}

//---------------------------------------------------------------------------
inline bool Transition::isFork() const
{
  return ( postset->count() > 1 );
}

//---------------------------------------------------------------------------
inline bool Transition::isJoin() const
{
  return ( preset->count() > 1 );
}

//---------------------------------------------------------------------------
inline int Transition::getPresetCount() const
{
  return preset->count();
}

//---------------------------------------------------------------------------
inline void Transition::setPreset(PlaceArray* _preset)
{
  preset = _preset;
}

//---------------------------------------------------------------------------
inline PlaceArray* Transition::getPreset() const
{
  return preset;
}

//---------------------------------------------------------------------------
inline Place* Transition::getPreset(int position) const
{
  return (*preset)[position];
}

//---------------------------------------------------------------------------
inline void Transition::addPreset(Place* place)
{
  preset->add(place);
}

//---------------------------------------------------------------------------
inline void Transition::replacePreset(Place* place, int position)
{
  (*preset)[position] = place;
}

//---------------------------------------------------------------------------
inline void Transition::removePreset(Place* place)
{
  preset->remove(place, false);
}

//---------------------------------------------------------------------------
inline int Transition::locatePreset(Place* place) const
{
  return preset->locate(place);
}

//---------------------------------------------------------------------------
inline int Transition::getPostsetCount() const
{
  return postset->count();
}

//---------------------------------------------------------------------------
inline void Transition::setPostset(PlaceArray* _postset)
{
  postset = _postset;
}

//---------------------------------------------------------------------------
inline PlaceArray* Transition::getPostset() const
{
  return postset;
}

//---------------------------------------------------------------------------
inline Place* Transition::getPostset(int position) const
{
  return (*postset)[position];
}

//---------------------------------------------------------------------------
inline void Transition::addPostset(Place* place)
{
  postset->add(place);
}

//---------------------------------------------------------------------------
inline void Transition::replacePostset(Place* place, int position)
{
  (*postset)[position] = place;
}

//---------------------------------------------------------------------------
inline void Transition::removePostset(Place* place)
{
  postset->remove(place, false);
}

//---------------------------------------------------------------------------
inline int Transition::locatePostset(Place* place) const
{
  return postset->locate(place);
}

//---------------------------------------------------------------------------
inline int Transition::getReadsetCount() const
{
  return readset->count();
}

//---------------------------------------------------------------------------
inline void Transition::setReadset(PlaceArray* _readset)
{
  readset = _readset;
}

//---------------------------------------------------------------------------
inline PlaceArray* Transition::getReadset() const
{
  return readset;
}

//---------------------------------------------------------------------------
inline Place* Transition::getReadset(int position) const
{
  return (*readset)[position];
}

//---------------------------------------------------------------------------
inline void Transition::addReadset(Place* place)
{
  readset->add(place);
}

//---------------------------------------------------------------------------
inline void Transition::replaceReadset(Place* place, int position)
{
  (*readset)[position] = place;
}

//---------------------------------------------------------------------------
inline void Transition::removeReadset(Place* place)
{
  readset->remove(place, false);
}

//---------------------------------------------------------------------------
inline int Transition::locateReadset(Place* place) const
{
  return readset->locate(place);
}

#endif
